/*
 * MODULE
 *
 *   init.c
 *
 * DESCRIPTION
 *
 *  Hardware Initialization Routines
 *
 * PLATFORM
 *
 *  Olimex LPC-2478-STK
 *
 * HISTORY
 *
 *   2012/11/21: John Hatch:
 *     Converted to use iolpc2470_gcc.h header
 *   2007/10/31: Mitch Ishihara:
 *     changed PINSEL10 setting to just write the value instead
 *       of read-modify-write.
 *   2007/10/01: Mitch Ishihara:
 *     split init routines from main.c
 *     changed MEMMAP to USERRAMMODE to run completely from RAM
 *   2007/09/12: Andrew N. Sloss:
 *     convert the main.c to use VOLATILExx(<address>) format
 *     added 1st draft comments
 *   2007/08/21: Andrew N. Sloss:
 *     inspired from James Lynch's example code demo2106_blink_flash.
 *
 * NOTES
 *
 *  Taken from the LPC2378 User Manual:
 *
 *   5.14 PLL setup sequence
 *
 *   The following sequence must be followed step by step in order to
 *   have the PLL initialized an running:
 *
 *   1. Disconnect the PLL with one feed sequence if PLL is already
 *      connected.
 *   2. Disable the PLL with one feed sequence.
 *   3. Change the CPU Clock Divider setting to speed up operation without
 *      the PLL, if desired.
 *   4. Write to the Clock Source Selection Control register to change the
 *      clock source.
 *   5. Write to the PLLCFG and make it effective with one feed sequence.
 *      The PLLCFG can only be updated when the PLL is disabled.
 *   6. Enable the PLL with one feed sequence.
 *   7. Change the CPU Clock Divider setting for the operation with the
 *      PLL. It's critical to do this before connecting the PLL.
 *   8. Wait for the PLL to achieve lock by monitoring the PLOCK bit in the
 *      PLLSTAT register, or using the PLOCK interrupt, or wait for a fixed
 *      time when the input clock to PLL is slow (i.e. 32 kHz). The value of
 *      PLOCK may not be stable when the PLL reference frequency (FREF, the
 *      frequency of REFCLK, which is equal to the PLL input frequency divided
 *      by the pre-divider value) is less than 100 kHz or greater than 20
 *      MHz. In these cases, the PLL may be assumed to be stable after a
 *      start-up time has passed. This time is 500 us when FREF is
 *      greater than 400 kHz and 200 / FREF seconds when FREF is less than
 *      400kHz.
 *   9. Connect the PLL with one feed sequence.
 *
 *   It's very important not to merge any steps above. For example, don't
 *   update the PLLCFG and enable the PLL simultaneously with the same feed
 *   sequence.
 *
 */

#include "io_macros.h"
#include "iolpc2470_gcc.h"
#include "pins.h"
#include "init.h"

//
//  InitPLL
//
//  This init is specific to the 2300 family of devices.
//
//  Initialize PLL (Configured for a 12.000 MHz crystal) to
//  boost processor clock to 72.000 MHz
//
static void InitPLL(void)
{
    // [1] Check if PLL connected, disconnect if yes.

    if (PLLSTAT & PLLSTAT_PLLC)
    {
        PLLCON = PLLCON_PLLE;
        // Enable PLL, disconnected ( PLLC = 0)
        PLLFEED = 0xaa;
        PLLFEED = 0x55;
    }

    // [2] Disable the PLL once it has been disconnected.

    PLLCON  = 0;
    PLLFEED = 0xaa;
    PLLFEED = 0x55;

    /* Enable the main oscillator */

    SCS |= SCS_OSCEN;

    // Wait until main OSC is usable

    do {} while ((SCS & SCS_OSCSTAT) == 0);

    // Select the main osc as the PLL clk source.

    CLKSRCSEL = CLKSRC_MAIN_OSC;

	// Set multiplier and divider values.
	//
	//  PLLCFG = ((N - 1) << 16) + (M - 1)
	//
	//  F_cco = (2 * M *F_in)/N
	//      M = 12, N = 1, F_in = 12.000 MHz -> F_cco = 288.000 MHz
    
    PLLCFG = PLLCFG_VALUE;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;
    
    // Enable the PLL.
    
    PLLCON  = PLLCON_PLLE;
    PLLFEED = 0xAA;
    PLLFEED = 0x55;
    
    // Divide F_cco down to get the CCLK output. (288 / 6 = 48)
    
    CCLKCFG = CCLKCFG_VALUE;
    
    // Wait for the PLL to lock to set frequency
    
    do {} while((PLLSTAT & PLLSTAT_PLOCK) == 0);
    
    do {} while((PLLSTAT & 0x00FF7FFF) != PLLCFG_VALUE);
    
    // Enable and connect the PLL as the clock source
    
    PLLCON = (PLLCON_PLLE | PLLCON_PLLC);
    PLLFEED = 0xAA;
    PLLFEED = 0x55;
    
    // Check connect bit status and wait for connection.
    
    do {} while((PLLSTAT & PLLSTAT_PLLC) == 0);
}

//
//  InitMAM
//
//  Eenable the MAM accelerator and sets number of clocks used for
//  flash memory fetch.
//
static void InitMAM(void)
{
    MAMCR = MAMCR_OFF;
    MAMTIM = MAMTIM_3_CLK;
    MAMCR = MAMCR_PART;
}

//
//  InitPCLK
//
//  Set the pclk for each peripheral. The default PDIV_4, but we do that
//  explicitly here to make it easier to change them later.
//
#define PDIV_1             0x1
#define PDIV_2             0x2
#define PDIV_4             0x0
#define PDIV_6             0x3
#define PDIV_8             0x3

#define SET_PCLK(per, div) ((div) << (per))

static void InitPCLK(void)
{
//    PCLKSEL0 =
//    (
//          SET_PCLK (PCLK_WDT,     PDIV_4)
//        | SET_PCLK (PCLK_TIMER0,  PDIV_4)
//        | SET_PCLK (PCLK_TIMER1,  PDIV_4)
//        | SET_PCLK (PCLK_UART0,   PDIV_1)
//        | SET_PCLK (PCLK_UART1,   PDIV_1)
//        | SET_PCLK (PCLK_PWM1,    PDIV_4)
//        | SET_PCLK (PCLK_I2C0,    PDIV_4)
//        | SET_PCLK (PCLK_SPI,     PDIV_4)
//        | SET_PCLK (PCLK_RTC,     PDIV_4)
//        | SET_PCLK (PCLK_SSP1,    PDIV_4)
//        | SET_PCLK (PCLK_DAC,     PDIV_4)
//        | SET_PCLK (PCLK_ADC,     PDIV_4)
//        | SET_PCLK (PCLK_CAN1,    PDIV_4)
//        | SET_PCLK (PCLK_CAN2,    PDIV_4)
//        | SET_PCLK (PCLK_ACF,     PDIV_4)
//    );

    PCLKSEL0_bit.PCLK_WDT = PDIV_4;
    PCLKSEL0_bit.PCLK_TIMER0 = PDIV_4;
    PCLKSEL0_bit.PCLK_TIMER1 = PDIV_4;
    PCLKSEL0_bit.PCLK_UART0 = PDIV_1;
    PCLKSEL0_bit.PCLK_UART1 = PDIV_1;
    PCLKSEL0_bit.PCLK_PWM1 = PDIV_4;
    PCLKSEL0_bit.PCLK_I2C0 = PDIV_4;
    PCLKSEL0_bit.PCLK_SPI = PDIV_4;
    PCLKSEL0_bit.PCLK_RTC = PDIV_4;
    PCLKSEL0_bit.PCLK_SSP1 = PDIV_4;
    PCLKSEL0_bit.PCLK_DAC = PDIV_4;
    PCLKSEL0_bit.PCLK_ADC = PDIV_4;
    PCLKSEL0_bit.PCLK_CAN1 = PDIV_4;
    PCLKSEL0_bit.PCLK_CAN2 = PDIV_4;
    PCLKSEL0_bit.PCLK_ACF = PDIV_4;


//    PCLKSEL1 =
//    (
//          SET_PCLK (PCLK_BAT_RAM, PDIV_4)
//        | SET_PCLK (PCLK_GPIO,    PDIV_4)
//        | SET_PCLK (PCLK_PCB,     PDIV_4)
//        | SET_PCLK (PCLK_I2C1,    PDIV_4)
//        | SET_PCLK (PCLK_SSP0,    PDIV_4)
//        | SET_PCLK (PCLK_TIMER2,  PDIV_4)
//        | SET_PCLK (PCLK_TIMER3,  PDIV_4)
//        | SET_PCLK (PCLK_UART2,   PDIV_1)
//        | SET_PCLK (PCLK_UART3,   PDIV_1)
//        | SET_PCLK (PCLK_I2C2,    PDIV_4)
//        | SET_PCLK (PCLK_I2S,     PDIV_4)
//        | SET_PCLK (PCLK_MCI,     PDIV_4)
//        | SET_PCLK (PCLK_SYSCON,  PDIV_4)
//    );

    PCLKSEL1_bit.PCLK_BAT_RAM = PDIV_4;
    PCLKSEL1_bit.PCLK_GPIO = PDIV_4;
    PCLKSEL1_bit.PCLK_PCB = PDIV_4;
    PCLKSEL1_bit.PCLK_I2C1 = PDIV_4;
    PCLKSEL1_bit.PCLK_SSP0 = PDIV_4;
    PCLKSEL1_bit.PCLK_TIMER2 = PDIV_4;
    PCLKSEL1_bit.PCLK_TIMER3 = PDIV_4;
    PCLKSEL1_bit.PCLK_UART2 = PDIV_4;
    PCLKSEL1_bit.PCLK_UART3 = PDIV_1;
    PCLKSEL1_bit.PCLK_I2C2 = PDIV_4;
    PCLKSEL1_bit.PCLK_I2S = PDIV_4;
    PCLKSEL1_bit.PCLK_MCI = PDIV_4;
    PCLKSEL1_bit.PCLK_SYSCON = PDIV_4;

}

//
//  pinConfigurator
//
//  Setups all the pins of the LPC2378
//
void PinConfigurator(void)
{
    uint32_t n;
    uint32_t i;
    uint32_t pattern;

    n = 0;

    while (mcuPins[n].pinSelect!=0)
    {
        pattern = 0x00000000;

        for (i=0; i<16; i++)
        {
            pattern |= (mcuPins[n].pin[i].multiplex << (mcuPins[n].pin[i].position*2));
        }

        // set the complete world
        mcuPins[n].pinSelect = pattern;
        n++;
    }

    // Note: Complete by setting the ETM to be disabled on PINSEL10

    // Disable ETM, mask reserved bit to 0
    PINSEL10 = ~(BIT(3)) & PINSEL10_MASK;
}

//
//  InitGPIO
//
//  Sets up the GPIO pins
//
static void InitGPIO(void)
{
    PinConfigurator();

	// NOTE:
	//
	//  SCS controls whether the LPC2378 is set to use the legancy registers or
	//  the new fast GPIO control registers.

    // Fast GPIO / new registers
    SCS |= BIT(0);
}

//
//  initUART0
//
//  Sets up UART0 to be 38.4 K Baud, 8 data bits, no parity, and no flow control
//
void InitUART0(uint16_t baud, uint8_t mode, uint8_t fmode)
{
    U0LCR = ULCR_DLAB_ENABLE;
    U0DLL = (uint8_t) baud;
    U0DLM = (uint8_t)(baud >> 8);
    U0LCR = (mode & ~ULCR_DLAB_ENABLE);
    U0FCR = fmode;
}

//
//   InitHardware
//
//   Initializes all the hardware components for the UART0 to function.
//
void InitHardware(void)
{
    // intialize main hardware components
    
    InitPLL();
    InitMAM();
    InitPCLK ();
    InitGPIO();
    
    // intialize specific hardware components for UART0
    InitUART0(UART_BAUD(HOST_BAUD_U0), UART_8N1,UART_FIFO_OFF);
    
    // Turn off MCIPWR SD LED (near bottom left corner of LCD)
    FIO1DIR |= 1<<5;
    FIO1CLR |= 1<<5;
    
    // MEMMAP Choices are:
    // BOOTLOADERMODE      0x00
    // USERFLASHMODE       0x01
    // USERRAMMODE         0x02
    // EXTERNALMEMORYMODE  0x03

    MEMMAP = USERFLASHMODE;
}


